/* asm.c - 6502 assembler (horrible unfinished hackofathing)
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define _GNU_SOURCE

#ifdef DEBUG
#define DEBUG_LABELS     0
#define DEBUG_BRANCHING  0
#endif


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xstring.h"
#include "xlib/xlist.h"
#include "xlib/xfile.h"
#include "xlib/xgetopt.h"

#include "opcodes.h"


#define VERSION              "1.0"
#define AUTHOR_NAME          "Tim Musson"
#define AUTHOR_EMAIL         "trmusson@gmail.com"

#define FNAME_OUT            "out"
#define FNAME_EXT_OBJ        "obj"
#define FNAME_EXT_HEX        "hex"
#define FNAME_EXT_ASM        "asm"
#define FNAME_EXT_SYM        "sym"

#define CHARS_UPPER          "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define CHARS_LOWER          "abcdefghijklmnopqrstuvwxyz"
#define CHARS_BIN            "01"
#define CHARS_DEC            "0123456789"
#define CHARS_HEX            CHARS_DEC "ABCDEFabcdef"
#define CHARS_LABEL          CHARS_UPPER CHARS_LOWER CHARS_DEC "_."

#define CHAR_COMMENT         ';'
#define CHAR_EQU             '='
#define CHAR_IMMEDIATE       '#'
#define CHAR_LSB             '<'
#define CHAR_MSB             '>'
#define CHAR_LBRACKET        '('
#define CHAR_RBRACKET        ')'

#define STR_PC               "*"
#define STR_ACCUMULATOR      "A"
#define STR_LO               "LO"
#define STR_HI               "HI"
#define STR_LSB              "<"
#define STR_MSB              ">"
#define STR_LBRACKET         "("
#define STR_RBRACKET         ")"

#define LABEL_TERMINATOR     ":"

#define SEPARATOR_ADDR       ": "
#define SEPARATOR_COL        " "

#define DIRECTIVE_BLKB       ".blkb"
#define DIRECTIVE_BLKW       ".blkw"
#define DIRECTIVE_BYTE       ".byte"
#define DIRECTIVE_END        ".end"
#define DIRECTIVE_EQU        ".equ"
#define DIRECTIVE_INCLUDE    ".include"
#define DIRECTIVE_ORG        ".org"
#define DIRECTIVE_WORD       ".word"

#define FMT_BYTE             "%02X"
#define FMT_WORD             "%04X"

#define TOKEN(x)                ((Token*)x->data)
#define TOK_TYPE(x)             (TOKEN(x)->type)
#define TOK_TEXT(x)             (TOKEN(x)->s)
#define TOK_IS_STRING(this)     (TOK_TYPE(this) == TOK_DATA \
                                 && *TOK_TEXT(this) == '"')

#define IS_BLOCK(instr)         ((instr)->block_size > 0)

#define OP_GET_CODE(id, mode)   opcode[id].code[mode]
#define OP_GET_MNEMONIC(id)     opcode[id].mnemonic
#define OP_GET_DEFAULT_MODE(id) opcode[id].default_mode

#define HAVE_OPCODE(instr)      ((instr)->opcode_id >= 0)

#define BYTE_SIZE               8
#define BYTE_MAX                256    /* 2 ^ BYTE_SIZE       */
#define WORD_MAX                65536  /* 2 ^ (BYTE_SIZE * 2) */

#define WORD_INC(w)             word_set (w, word_to_integer (w) + 1)
#define WORD_DEC(w)             word_set (w, word_to_integer (w) - 1)
#define WORD_ADD_BYTE(w, b)     word_set (w, word_to_integer (w) + b)
#define WORD_SUB_BYTE(w, b)     word_set (w, word_to_integer (w) - b)
#define WORD_ADD_WORD(w1, w2)   word_set (w1, word_to_integer (w1) + word_to_integer (w2))
#define WORD_EQ(w1, w2)         ((w1)->lsb == (w2)->lsb && (w1)->msb == (w2)->msb)


typedef unsigned char Byte;


typedef struct {
	Byte lsb;
	Byte msb;
} Word;


enum {
	OUTPUT_OBJ,
	OUTPUT_HEX,
	OUTPUT_ASM,
	OUTPUT_SYM
};


typedef enum {
	DATA_TYPE_BYTE,
	DATA_TYPE_WORD
} DataType;


typedef enum {
	TOK_NONE,
	TOK_OPCODE,
	TOK_DATA,
	TOK_LABEL,
	TOK_DIRECTIVE,
	TOK_SYNTAX
} TokenType;


typedef enum {
	AM_IMMEDIATE,
	AM_DIRECT,
	AM_DIRECT_X,
	AM_DIRECT_Y,
	AM_ZERO_PAGE,
	AM_ZERO_PAGE_X,
	AM_ZERO_PAGE_Y,
	AM_INDIRECT,
	AM_INDIRECT_X,
	AM_INDIRECT_Y,
	AM_IMPLIED,
	AM_RELATIVE,
	AM_ACCUMULATOR,

	N_ADDR_MODES

} AddrMode;

#define AM_NONE    N_ADDR_MODES


typedef struct {
	char *mnemonic;          /* jmp, lda, rts, sta, etc. */
	int  code[N_ADDR_MODES]; /* Code according to addressing mode */
} Opcode;


typedef struct {
	char    *name;
	Word    data;
	xbool_t assigned;

	/* Which source file and line number this label came from */

	char *fname;
	int  line_num;
} Label;


typedef struct {
	Word     addr;     /* Address (program counter value) of this instruction */
	int      opcode_id;/* Index to opcode[], or -1 if no CPU operation        */
	AddrMode mode;     /* If opcode_id != -1, address mode for this operation */

	Word data;       /* Data to work on; the operand if opcode_id != -1       */
	                 /* XXX: replace instruction->data with instruction->list */

	DataType dtype;       /* Whether data's a full word or 1 byte (data.lsb)  */
	xlist_t  *list;       /* List of Word, used for .byte and .word values    */
	int      block_size;  /* If non-zero, make a block of data */
	Label    *label;      /* If not NULL, copy data from label */
	xbool_t  immediate_msb;  /* Req. MSB of immediate mode operand, eg. #>FOO */
	xlist_t  *expr;       /* e.g. + 5 */

	/* Which source file and line number this label came from */
	/* XXX: Wait... can't this come from label->fname, label->line_num? */

	char     *fname;
	int      line_num;
} Instruction;


typedef struct {
	char *fname;
	FILE *f_in;
	int  parent_line_num;
} Include;


typedef struct {
	char *s;
	TokenType type;
} Token;


typedef struct {
	char    *fname_in;
	char    *fname_out;
	FILE    *f_out;
	int     line_num;
	Word    pc;            /* Program Counter */
	xlist_t *labels;       /* List of Label */
	xlist_t *instructions; /* List of Instruction */
	xlist_t *includes;     /* List of currently open files */
	xlist_t *errors;       /* List of error strings to print on exit */
	Label   *label_pc;     /* Pointer to special label "*" */
} Parser;

 
static Opcode opcode[] = {

	/* See opcodes.h */

	{"adc", {OP_ADC, OP_ADC_A, OP_ADC_AX, OP_ADC_AY, OP_ADC_Z, OP_ADC_ZX, OP_ADC_ZY, OP_ADC_I, OP_ADC_IX, OP_ADC_IY, -1,         -1,       -1        }},
	{"and", {OP_AND, OP_AND_A, OP_AND_AX, OP_AND_AY, OP_AND_Z, OP_AND_ZX, OP_AND_ZY, OP_AND_I, OP_AND_IX, OP_AND_IY, -1,         -1,       -1        }},
	{"asl", {-1,     OP_ASL_A, OP_ASL_AX, -1,        OP_ASL_Z, OP_ASL_ZX, -1,        -1,       -1,        -1,        -1,         -1,       OP_ASL_ACC}},
	{"bit", {-1,     OP_BIT_A, -1,        -1,        OP_BIT_Z, -1,        -1,        -1,       -1,        -1,        -1,         -1,       -1        }},
	{"bcc", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BCC_R, -1        }},
	{"bcs", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BCS_R, -1        }},
	{"beq", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BEQ_R, -1        }},
	{"bmi", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BMI_R, -1        }},
	{"bne", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BNE_R, -1        }},
	{"bpl", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BPL_R, -1        }},
	{"bvc", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BVC_R, -1        }},
	{"bvs", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         OP_BVS_R, -1        }},
	{"brk", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_BRK_IMP, -1,       -1        }},
	{"clc", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_CLC_IMP, -1,       -1        }},
	{"cld", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_CLD_IMP, -1,       -1        }},
	{"cli", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_CLI_IMP, -1,       -1        }},
	{"clv", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_CLV_IMP, -1,       -1        }},
	{"cmp", {OP_CMP, OP_CMP_A, OP_CMP_AX, OP_CMP_AY, OP_CMP_Z, OP_CMP_ZX, OP_CMP_ZY, OP_CMP_I, OP_CMP_IX, OP_CMP_IY, -1,         -1,       -1        }},
	{"cpx", {OP_CPX, OP_CPX_A, -1,        -1,        OP_CPX_Z, -1,        -1,        -1,       -1,        -1,        -1,         -1,       -1        }},
	{"cpy", {OP_CPY, OP_CPY_A, -1,        -1,        OP_CPY_Z, -1,        -1,        -1,       -1,        -1,        -1,         -1,       -1        }},
	{"dec", {OP_DEC, OP_DEC_A, OP_DEC_AX, OP_DEC_AY, OP_DEC_Z, OP_DEC_ZX, OP_DEC_ZY, OP_DEC_I, OP_DEC_IX, OP_DEC_IY, -1,         -1,       -1        }},
	{"dex", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_DEX_IMP, -1,       -1        }},
	{"dey", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_DEY_IMP, -1,       -1        }},
	{"eor", {OP_EOR, OP_EOR_A, OP_EOR_AX, OP_EOR_AY, OP_EOR_Z, OP_EOR_ZX, OP_EOR_ZY, OP_EOR_I, OP_EOR_IX, OP_EOR_IY, -1,         -1,       -1        }},
	{"inc", {OP_INC, OP_INC_A, OP_INC_AX, OP_INC_AY, OP_INC_Z, OP_INC_ZX, OP_INC_ZY, OP_INC_I, OP_INC_IX, OP_INC_IY, -1,         -1,       -1        }},
	{"inx", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_INX_IMP, -1,       -1        }},
	{"iny", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_INY_IMP, -1,       -1        }},
	{"jmp", {-1,     OP_JMP_A, -1,        -1,        -1,       -1,        -1,        OP_JMP_I, -1,        -1,        -1,         -1,       -1        }},
	{"jsr", {-1,     OP_JSR_A, -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        -1,         -1,       -1        }},
	{"lda", {OP_LDA, OP_LDA_A, OP_LDA_AX, OP_LDA_AY, OP_LDA_Z, OP_LDA_ZX, OP_LDA_ZY, OP_LDA_I, OP_LDA_IX, OP_LDA_IY, -1,         -1,       -1        }},
	{"ldx", {OP_LDX, OP_LDX_A, OP_LDX_AX, OP_LDX_AY, OP_LDX_Z, OP_LDX_ZX, OP_LDX_ZY, OP_LDX_I, OP_LDX_IX, OP_LDX_IY, -1,         -1,       -1        }},
	{"ldy", {OP_LDY, OP_LDY_A, OP_LDY_AX, OP_LDY_AY, OP_LDY_Z, OP_LDY_ZX, OP_LDY_ZY, OP_LDY_I, OP_LDY_IX, OP_LDY_IY, -1,         -1,       -1        }},
	{"lsr", {OP_LSR, OP_LSR_A, OP_LSR_AX, OP_LSR_AY, OP_LSR_Z, OP_LSR_ZX, OP_LSR_ZY, OP_LSR_I, OP_LSR_IX, OP_LSR_IY, -1,         -1,       OP_LSR_ACC}},
	{"nop", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_NOP_IMP, -1,       -1        }},
	{"ora", {OP_ORA, OP_ORA_A, OP_ORA_AX, OP_ORA_AY, OP_ORA_Z, OP_ORA_ZX, OP_ORA_ZY, OP_ORA_I, OP_ORA_IX, OP_ORA_IY, -1,         -1,       -1        }},
	{"pha", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_PHA_IMP, -1,       -1        }},
	{"pla", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_PLA_IMP, -1,       -1        }},
	{"php", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_PHP_IMP, -1,       -1        }},
	{"plp", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_PLP_IMP, -1,       -1        }},
	{"rol", {-1,     OP_ROL_A, OP_ROL_AX, -1,        OP_ROL_Z, OP_ROL_ZX, -1,        -1,       -1,        -1,        -1,         -1,       OP_ROL_ACC}},
	{"ror", {-1,     OP_ROR_A, OP_ROR_AX, -1,        OP_ROR_Z, OP_ROR_ZX, -1,        -1,       -1,        -1,        -1,         -1,       OP_ROR_ACC}},
	{"rts", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_RTS_IMP, -1,       -1        }},
	{"rti", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_RTI_IMP, -1,       -1        }},
	{"sbc", {OP_SBC, OP_SBC_A, OP_SBC_AX, OP_SBC_AY, OP_SBC_Z, OP_SBC_ZX, OP_SBC_ZY, OP_SBC_I, OP_SBC_IX, OP_SBC_IY, -1,         -1,       -1        }},
	{"sec", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_SEC_IMP, -1,       -1        }},
	{"sed", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_SED_IMP, -1,       -1        }},
	{"sei", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_SEI_IMP, -1,       -1        }},
	{"sta", {OP_STA, OP_STA_A, OP_STA_AX, OP_STA_AY, OP_STA_Z, OP_STA_ZX, OP_STA_ZY, OP_STA_I, OP_STA_IX, OP_STA_IY, -1,         -1,       -1        }},
	{"stx", {OP_STX, OP_STX_A, OP_STX_AX, OP_STX_AY, OP_STX_Z, OP_STX_ZX, OP_STX_ZY, OP_STX_I, OP_STX_IX, OP_STX_IY, -1,         -1,       -1        }},
	{"sty", {OP_STY, OP_STY_A, OP_STY_AX, OP_STY_AY, OP_STY_Z, OP_STY_ZX, OP_STY_ZY, OP_STY_I, OP_STY_IX, OP_STY_IY, -1,         -1,       -1        }},
	{"tax", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TAX_IMP, -1,       -1        }},
	{"tay", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TAY_IMP, -1,       -1        }},
	{"txa", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TXA_IMP, -1,       -1        }},
	{"txs", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TXS_IMP, -1,       -1        }},
	{"tsx", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TSX_IMP, -1,       -1        }},
	{"tya", {-1,     -1,       -1,        -1,        -1,       -1,        -1,        -1,       -1,        -1,        OP_TYA_IMP, -1,       -1        }},

};


static XGetOption option[] = {
	{3, FALSE, "obj"},
	{3, FALSE, "hex"},
	{3, FALSE, "asm"},
	{3, FALSE, "symtab"},
	{1, FALSE, "counter"},
	{2, TRUE,  "origin"},
	{2, TRUE,  "saddress"},
	{2, TRUE,  "scolumn"},
	{1, TRUE,  "output"},
	{1, FALSE, "lower"},
	{1, FALSE, "upper"},
	{1, TRUE,  "width"},
	{1, FALSE, "quiet"},
	{1, FALSE, "version"},
	{1, FALSE, "help"}
};


enum {
	OPT_OBJ,
	OPT_HEX,
	OPT_ASM,
	OPT_SYM,
	OPT_COUNTER,
	OPT_ORIGIN,
	OPT_SEPARATOR_ADDR,
	OPT_SEPARATOR_COL,
	OPT_OUTPUT,
	OPT_LOWER,
	OPT_UPPER,
	OPT_WIDTH,
	OPT_QUIET,
	OPT_VERSION,
	OPT_HELP,

	N_OPTS
};


#define DEFAULT_OUTPUT_MODE  OUTPUT_HEX
#define DEFAULT_UPPERCASE    TRUE
#define DEFAULT_DUMP_WIDTH   16


static char *prog = NULL;
static Parser *parser = NULL;

static int output_mode = DEFAULT_OUTPUT_MODE;
static int dump_width = DEFAULT_DUMP_WIDTH;
static xbool_t uppercase_output = DEFAULT_UPPERCASE;
static xbool_t show_counter = FALSE;
static xbool_t quiet = FALSE;
static char *sep_col = NULL;
static char *sep_addr = NULL;


static xlist_t* load_source (Parser *p, const char *fname);


static void
word_copy (Word *w_dest, Word *w_source)
{
	w_dest->lsb = w_source->lsb;
	w_dest->msb = w_source->msb;
}


static void
word_set (Word *w, int data)
{
	w->msb = (data >> 8) & 0xFF;
	w->lsb = data & 0xFF;
}


static int
word_to_integer (Word *w)
{
	return (int)(w->msb * BYTE_MAX) + w->lsb;
}


/* Complain and exit, hopefully with a useful message, file name,
   and line number.  */

static void
fatal (char *msg, const char *fname, int line_num)
{
	if (!msg) {
		msg = xstr_dup ("Syntax");
	}

	if (fname && line_num > 0) {
		fprintf (stderr, "%s: ERROR: %s (Line %d, %s)\n",
		         prog, msg, line_num, fname);
	}
	else if (fname) {
		fprintf (stderr, "%s: ERROR: %s (%s)\n", prog, msg, fname);
	}
	else {
		fprintf (stderr, "%s: ERROR: %s\n", prog, msg);
	}

	xfree (msg);

	exit (EXIT_FAILURE);
}


static void
string_data_error (Parser *p, const char *s)
{
	fatal (xstr_dup_printf ("Unable to parse value '%s'", s),
	                        p->fname_in, p->line_num);
}


/* Convert each occurrence of the sequence "\t" in s to '\t',
   shifting any following characters forward accordingly.  */

static char*
unescape_tabs (char *s)
{
	char *q, *ret;

	q = ret = s;

	while (*q != '\0') {

		if (*s == '\\' && *(s + 1) == 't') {
			*q = '\t';
			s++;
		}
		else {
			*q = *s;
		}
		q++;
		s++;
	}

	return ret;
}


/* Convert a string of ASCII digits to an integer.
   If the string begins with "0x, "0X", or "$", assume it's hexadecimal.  */

static int
parse_ascii_integer (const char *s)
{
	char tmp[6];

	if (strncasecmp (s, "0x", 2) == 0) {
		s++;
	}

	if (*s == '$' || *s == 'x' || *s == 'X') {
		s++;
		sscanf (s, "%4[0-9A-Fa-f]", tmp);
		return strtol (tmp, NULL, 16);
	}

	return atoi (s);
}


static AddrMode
opcode_default_address_mode (Opcode *opcode)
{
	/* If there's an direct version, that's the default. */
	if (opcode->code[AM_DIRECT] != -1) {
		return AM_DIRECT;
	}

	/* Otherwise, if there's a relative version, that's the default. */
	if (opcode->code[AM_RELATIVE] != -1) {
		return AM_RELATIVE;
	}

	/* Otherwise, assume AM_IMPLIED. */
	assert (opcode->code[AM_IMPLIED] != -1);

	return AM_IMPLIED;
}


static xbool_t
opcode_is_branch (Opcode *opcode)
{
	/* Branches are the only ones using relative addressing, so... */
	return opcode_default_address_mode (opcode) == AM_RELATIVE;
}


static int
opcode_id_from_mnemonic (const char *mnemonic)
{
	int i, n;

	n = sizeof (opcode) / sizeof (Opcode);

	for (i = 0; i < n; i++) {
		if (strcasecmp (opcode[i].mnemonic, mnemonic) == 0) {
			return i;
		}
	}

	return -1;
}


static xbool_t
op_is_shift (int opcode_id)
{
	int test = opcode[opcode_id].code[AM_DIRECT];
	return test == OP_ASL_A
	               || test == OP_LSR_A || test == OP_ROL_A || test == OP_ROR_A;
}


static Label*
label_lookup_by_name (Parser *p, const char *name)
{
	xlist_t *list;

	for (list = p->labels; list != NULL; list = list->next) {
		if (strcmp (((Label*)list->data)->name, name) == 0) {
			return list->data;
		}
	}

	return NULL;
}


/* Assign data to label.  */

static void
label_assign (Parser *p, Label *label, Word *data)
{
	assert (data != NULL);
#if DEBUG_LABELS
	fprintf (stderr, "%s: Assigning '%s' (%02X %02X (msb, lsb))\n",
	         prog, label->name, data->msb, data->lsb);
#endif
	label->data.msb = data->msb;
	label->data.lsb = data->lsb;
	label->assigned = TRUE;

	if (label == p->label_pc) {
		/* Copy to Program Counter. */
		p->pc.msb = label->data.msb;
		p->pc.lsb = label->data.lsb;
#if DEBUG_LABELS
		fprintf (stderr, "%s: Program Counter set to $%02X%02X\n",
		         prog, p->pc.msb, p->pc.lsb);
#endif
	}
}


/* Return the named label.
   If it doesn't already exist, create it.  */

static Label*
get_label (Parser *p, const char *name, Word *data)
{
	Label *label;

	if (strcasecmp (name, STR_ACCUMULATOR) == 0
	    || strcasecmp (name, STR_LO) == 0
	    || strcasecmp (name, STR_HI) == 0) {
		fatal (xstr_dup_printf ("Illegal use of reserved label '%s'", name),
		       p->fname_in, p->line_num);
	}

	if ((label = label_lookup_by_name (p, name)) != NULL) {

		/* If un-assigned, assign and return. */
		if (data != NULL && (label == p->label_pc || !label->assigned)) {
			label_assign (p, label, data);
			return label;
		}
		if (label != p->label_pc) {
			fprintf (stderr, "%s: WARNING: Redefinition of label '%s' in '%s', line %d\n",
			         prog, label->name, p->fname_in, p->line_num);
			fprintf (stderr, "%s: First defined in '%s', line %d\n",
			         prog, label->fname, label->line_num);
			if (data != NULL) {
				label_assign (p, label, data);
			}
		}
		return label;
	}

	/* Create it. */
	label = xnew0 (Label, 1);
	label->name = xstr_dup (name);
	label->fname = xstr_dup (p->fname_in);
	label->line_num = p->line_num;
	if (data != NULL) {
		label_assign (p, label, data);
	}
	xlist_append (&p->labels, label);

	return label;
}


static void
label_free (Label *label)
{
	xfree (label->name);
	xfree (label->fname);
	xfree (label);
}


/* If the label exists and has been assigned a value, copy that
   value into the given data structure. Return TRUE if successful.  */

static xbool_t
label_get_data (Parser *p, Label *label, Word *data)
{
	if (label == p->label_pc) {

		/* It's "*", so copy the program counter. */

		/* XXX: checkme */
#if 1
		word_set (data, word_to_integer (&p->pc) - 1);
#else
		data->msb = p->pc.msb;
		data->lsb = p->pc.lsb;
#endif
	}
	else {
		if (!label->assigned) {
			return FALSE;
		}
		data->msb = label->data.msb;
		data->lsb = label->data.lsb;
	}

	return TRUE;
}


static Instruction*
instruction_new (Parser *p, int id)
{
	Instruction *instr;

	instr = xnew0 (Instruction, 1);
	instr->opcode_id = id;
	instr->mode = AM_NONE;
	instr->fname = xstr_dup (p->fname_in);
	instr->line_num = p->line_num;

	return instr;
}


static void
instruction_free (Instruction *instr)
{
	xlist_t *list;

	for (list = instr->list; list != NULL; list = list->next) {
		xfree (list->data);
	}
	xlist_free (&instr->list);

	for (list = instr->expr; list != NULL; list = list->next) {
		xfree (list->data);
	}
	xlist_free (&instr->expr);

	xfree (instr->fname);
	xfree (instr);
}


static char
unescape_char (char c)
{
	switch (c) {
		case '0':
			c = '\0';
			break;
		case 'a':
			c = '\a';
			break;
		case 'b':
			c = '\b';
			break;
		case 'e':
			/* '\e' isn't standard C, so... */
			c = 0x1B;
			break;
		case 'f':
			c = '\f';
			break;
		case 'n':
			c = '\n';
			break;
		case 'r':
			c = '\r';
			break;
		case 's':
			/* Specific to this assembler; we can use either '\s or ' '. */
			c = ' ';
			break;
		case 't':
			c = '\t';
			break;
		case 'v':
			c = '\v';
			break;
	}
	return c;
}


/* Parse string s into data.  */

static xbool_t
parse_string_data (Parser *p, const char *s, Word *data, xbool_t *is_word)
{
	char tmp[7];
	int n;

	if (is_word != NULL) {
		*is_word = FALSE;
	}

	if (*s == '\'') {

		/* Expect an ASCII character, e.g. 'A or '\t, optionally
		   followed by a closing '. */

		char c;

		s++;
		c = *s;

		if (*s == '\\') {
			s++;
			c = unescape_char (*s);
		}

		data->lsb = c;
		/* XXX: wait... shouldn't we also set data->msb to zero? */

		if (*s != '\0' && *(s + 1) == '\'') {
			s++;
		}

	}
	else if (*s == '$' || strncasecmp (s, "0x", 2) == 0) {

		/* Expect a hexadecimal number up to 4 chars in length. */

		if (*s == '0') {
			s++;
		}
		s++;

		if ((n = sscanf (s, "%4[0-9A-Fa-f]", tmp)) != 1) {
			return FALSE;
		}

		word_set (data, strtol (tmp, NULL, 16));

	}
	else if (*s == '%') {

		/* Expect a binary number up to 16 chars in length. */

		s++;

		if (strlen (s) > 16) {
			return FALSE;
		}

		word_set (data, strtol (s, NULL, 2));

	}
#if 0
	else if (*s == '-' && isdigit (*(s + 1))) {

		/* Expect a negative integer up to 6 chars in length (incl '-'). */

		if ((n = sscanf (s, "-%5[0-9]", tmp)) != 1) {
			return FALSE;
		}

		word_set (data, 0 - atoi (tmp));

	}
#endif
	else if (isdigit (*s)) {

		/* Expect a positive integer up to 5 chars in length. */

		if ((n = sscanf (s, "%5[0-9]", tmp)) != 1) {
			return FALSE;
		}

		word_set (data, atoi (tmp));

	}
	else {

		/* Expect a label. */

		Label *label;

		label = label_lookup_by_name (p, s);

		if (label != NULL) {
			if (!label->assigned) {
				/* XXX: Maybe better to look up the data later, not
				   during parsing. */
				fatal (xstr_dup_printf ("Label '%s' was never assigned", s),
				       p->fname_in, p->line_num);
			}
			if (label == p->label_pc) {
				word_copy (data, &p->pc);
			}
			else {
				word_copy (data, &label->data);
			}
		}
		else {
			/* Register a new label. */
			label = get_label (p, s, NULL);
		}
	}

	if (is_word != NULL) {
		*is_word = word_to_integer (data) > 0xFF;
	}

	return TRUE;
}


static xbool_t
str_contains_char (const char *s, char c)
{
	for (; *s != '\0'; s++) {
		if (*s == c) {
			return TRUE;
		}
	}

	return FALSE;
}


static xbool_t
is_space (char c)
{
	return c == ' ' || c == '\t';
}


#if 0
/* XXX: This is a crappy hack to help parse negative
        integers in immediate mode. */
static xbool_t immediate = FALSE;
#endif

static char*
get_token (Parser *p, const char *text, int start, int *end, TokenType *type,
           int *length)
{
	const char *s, *expected, *tmp;
	char *token;


	*type = TOK_NONE;
	*length = 0;

	while (is_space (text[start])) {
		start++;
	}

	s = &text[start];

	if (*s == '\0' || *s == CHAR_COMMENT) {
		return NULL;
	}

	if (*s == '.') {
		/* Assembler directive begins. */
		*type = TOK_DIRECTIVE;
		expected = CHARS_LABEL;
		s++;
	}
	else if ((strncasecmp (s, STR_LO, 2) == 0
		      || strncasecmp (s, STR_HI, 2) == 0)
		     && is_space (*(s + 2))) {
		/* Syntax. */
		*type = TOK_SYNTAX;
		*end = start + 2;
		*length = 2;
		return xstr_ndup (*s == 'l' || *s == 'L' ? STR_LSB : STR_MSB, 1);
	}
	else if (str_contains_char (CHARS_UPPER CHARS_LOWER "_", *s)) {
		/* Found either a label or an opcode - assume
		   it's a label for now.  */
		*type = TOK_LABEL;
		expected = CHARS_LABEL;
	}
	else if (*s == '*') {
		/* The Program Counter label. */
		/* XXX: unless it's a multiplier in an expression! */
		*type = TOK_LABEL;
		*end = start + 1;
		*length = 1;
		return xstr_ndup (s, 1);
	}
	else if (*s == '\'') {
		/* Character data begins. */
		int i = 2;
		if (*(s + 1) == '\\') {
			i++;
			if (*(s + 3) == '\'') {
				i++;
			}
		}
		else if (*(s + 2) == '\'') {
			i++;
		}
		*type = TOK_DATA;
		*end = start + i;
		*length = i;
		return xstr_ndup (s, i);
	}
	else if (*s == '"') {
		/* String data begins. */
		xstring_t *xstr = xstring_new ("");
		*type = TOK_DATA;
		while (*s != '\0') {
			if (*s == '\\' && *(s + 1) != '0') {
				s++;
				xstring_append_c (xstr, unescape_char (*s));
			}
			else if (xstr->len > 0 && *s == '"') {
				break;
			}
			else {
				xstring_append_c (xstr, *s);
			}
			s++;
		}
		if (*s != '"') {
			fatal (xstr_dup ("Unterminated string"), p->fname_in, p->line_num);
		}
		s++;
		*length = s - &text[start];
		*end = start + *length;
		return xstring_free (xstr, FALSE);
	}
	else if (*s == '$' || strncasecmp (s, "0x", 2) == 0) {
		/* Hexadecimal value begins. */
		*type = TOK_DATA;
		expected = CHARS_HEX;
		if (*s == '0') {
			s++;
		}
		s++;
	}
	else if (*s == '%') {
		/* Binary value begins. */
		*type = TOK_DATA;
		expected = CHARS_BIN;
		s++;
	}
	else if (str_contains_char (CHARS_DEC, *s)) {
		/* Decimal value begins. */
		*type = TOK_DATA;
		expected = CHARS_DEC;
	}
	else if (*s == ':'
	         || *s == ','
	         || *s == CHAR_IMMEDIATE
	         || *s == CHAR_LSB
	         || *s == CHAR_MSB
	         || *s == CHAR_EQU
	         || *s == CHAR_LBRACKET
	         || *s == CHAR_RBRACKET
	         || *s == '+'
	         || (*s == '-')
	         || (*s == '*')
	         || *s == '/') {

		/* Syntax. */

		*type = (*s == CHAR_EQU ? TOK_DIRECTIVE : TOK_SYNTAX);
		*end = start + 1;
		*length = 1;
		return xstr_ndup (s, 1);
	}
	else {
		/* Unexpecter character. */
		fatal (NULL, p->fname_in, p->line_num);
	}

	tmp = s;

#if 0
	if (*s == '-' && immediate) {
		s++;
	}
#endif
	while (str_contains_char (expected, *s)) {
		s++;
	}
	/* Error if < 1 char skipped. */
	if (tmp == s) {
		fatal (NULL, p->fname_in, p->line_num);
	}

	*length = s - &text[start];
	token = *length > 0 ? xstr_ndup (&text[start], *length) : NULL;
	*end = start + *length;

	/* Distinguish opcodes from labels. */
	if (*type == TOK_LABEL) {
		int id;
		id = opcode_id_from_mnemonic (token);
		if (id != -1) {
			*type = TOK_OPCODE;
		}
#if 0
#if DEBUG_LABELS
		else {
			fprintf (stderr, ">> [%s]\n", token);
		}
#endif
#endif
	}

	return token;
}


static int
get_branch_offset (Parser *p, Instruction *instr, Label *label)
{
	Word data;
	int here, there;
	signed char offset;

	here = word_to_integer (&instr->addr);

	label_get_data (p, label, &data);
	there = word_to_integer (&data);

	/* offset = (there - here) - 2; */
	if (there > here) {
		offset = there - here;
		offset = offset - 2;
	}
	else if (there < here) {
		offset = - (here - there + 2);
	}
	else {
		/* XXX: complain? */
		offset = 0;
	}

#if DEBUG_BRANCHING
	fprintf (stderr,
	         "%s: get_branch_offset: PC=$%04X, %s=$%04X, offset=$%02X (%d); $%04X + $%02X = $%04X\n",
	         prog, here, label->name, there, offset, offset, here, offset,
	         here + offset);
#endif

	return offset;
}


/* Derive instr->data from instr->label. */

static void
instruction_data_from_label (Parser *p, Instruction *instr)
{
	Label *label;

	assert (instr->label != NULL);

	label = instr->label;

	if (!label->assigned) {
		fatal (xstr_dup_printf ("No value assigned to label '%s'",
		       label->name), instr->fname, instr->line_num);
	}

	assert (instr->opcode_id != -1);

	if (opcode_is_branch (&opcode[instr->opcode_id])) {
		/* This is a branch instruction, so calculate the offset. */
		word_set (&instr->data, get_branch_offset (p, instr, label));
	}
	else {
		/* Otherwise copy the data as-is. */
		label_get_data (p, label, &instr->data);
	}
}


static void
write_byte (Parser *p, Byte b, xbool_t prefix)
{
	if (prefix) {
		fprintf (p->f_out, "$");
	}
	fprintf (p->f_out, FMT_BYTE, b);
}


static void
write_word (Parser *p, Word *word, xbool_t prefix)
{
	if (prefix) {
		fprintf (p->f_out, "$");
	}
	fprintf (p->f_out, FMT_WORD, word_to_integer (word));
}


static void
write_char (Parser *p, char c)
{
	if (uppercase_output) {
		c = toupper (c);
	}

	fprintf (p->f_out, "%c", c);
}


static void
write_int (Parser *p, int n)
{
	fprintf (p->f_out, "%d", n);
}


static void
write_string (Parser *p, const char *s)
{
	char *q;

	if (uppercase_output) {
		q = xstr_dup_toupper (s);
		fprintf (p->f_out, "%s", q);
		xfree (q);
	}
	else {
		fprintf (p->f_out, "%s", s);
	}
}


static void
write_program_counter (Parser *p, Word *addr)
{
	if (!show_counter) {
		return;
	}

	write_word (p, addr, FALSE);
	write_string (p, sep_addr);
}


static void
dump_byte (Parser *p, int cell, Byte b)
{
	static int cell_prev = -1;
	Word addr;

	word_set (&addr, cell);

	if (cell_prev == -1) {
		write_program_counter (p, &addr);
		cell_prev = cell;
	}
	else {
		if (cell != cell_prev + 1) {
			write_char (p, '\n');
			write_program_counter (p, &addr);
		}
		cell_prev = cell;
	}
	write_byte (p, b, FALSE);
	write_string (p, sep_col);
}


static void
write_instruction_obj (Parser *p, Instruction *instr)
{
	if (HAVE_OPCODE(instr)) {
		fprintf (p->f_out, "%c", OP_GET_CODE(instr->opcode_id, instr->mode));
	}

	if (IS_BLOCK(instr)) {
		int i;
		for (i = 0; i < instr->block_size; i++) {
			fprintf (p->f_out, "%c", instr->data.lsb);
			if (instr->dtype == DATA_TYPE_WORD) {
				fprintf (p->f_out, "%c", instr->data.msb);
			}
		}
	}
	else if (!HAVE_OPCODE(instr)
		     || (instr->mode != AM_IMPLIED && instr->mode != AM_ACCUMULATOR)) {

		if (instr->label) {
			instruction_data_from_label (p, instr);
		}

		/* XXX */
		if (instr->expr) {
			xlist_t *e = instr->expr;
			char c = *((char*)e->data);
			int i; /* = atoi ((char*)instr->expr->next->data); */
			Word data;
			if (!parse_string_data (p, instr->expr->next->data, &data, NULL)) {
				/* XXX */
				fatal (NULL, instr->fname, instr->line_num);
			}
			i = word_to_integer (&data);
			switch (c) {
			case '+':
				word_set (&instr->data, word_to_integer (&instr->data) + i);
				break;
			case '-':
				word_set (&instr->data, word_to_integer (&instr->data) - i);
				break;
			case '*':
				word_set (&instr->data, word_to_integer (&instr->data) * i);
				break;
			case '/':
				word_set (&instr->data, word_to_integer (&instr->data) / i);
				break;
			default:
				fatal (NULL, instr->fname, instr->line_num);
			}
		}

		if (instr->list) {
			xlist_t *tmp;
			for (tmp = instr->list; tmp != NULL; tmp = tmp->next) {
				Word *data = tmp->data;
				/* XXX: checkme - if DATA_TYPE_WORD, write LSB or MSB first? */
				fprintf (p->f_out, "%c", data->lsb);
				if (instr->dtype == DATA_TYPE_WORD) {
					fprintf (p->f_out, "%c", data->msb);
				}
			}
		}
		if (!HAVE_OPCODE(instr)
		    || instr->mode == AM_IMMEDIATE
		    || instr->mode == AM_RELATIVE
		    || instr->mode == AM_ZERO_PAGE) {
			if (instr->immediate_msb) {
				fprintf (p->f_out, "%c", instr->data.msb);
			}
			else {
				fprintf (p->f_out, "%c", instr->data.lsb);
			}
		}
		else {
			fprintf (p->f_out, "%c%c", instr->data.lsb, instr->data.msb);
		}
	}
}


static int
write_instruction_hex (Parser *p, Instruction *instr, int col, xbool_t is_last)
{
	int cell = word_to_integer (&instr->addr);

	if (HAVE_OPCODE(instr)) {
		dump_byte (p, cell, OP_GET_CODE(instr->opcode_id, instr->mode));
		cell++;
	}

	if (IS_BLOCK(instr)) {
		int i;
		for (i = 0; i < instr->block_size; i++) {
			dump_byte (p, cell, instr->data.lsb);
			cell++;
			WORD_INC (&instr->addr);
			if (instr->dtype == DATA_TYPE_WORD) {
				dump_byte (p, cell, instr->data.msb);
				cell++;
				WORD_INC (&instr->addr);
			}
		}
	}
	else if (!HAVE_OPCODE(instr)
	         || (instr->mode != AM_IMPLIED && instr->mode != AM_ACCUMULATOR)) {

		if (instr->label) {
			instruction_data_from_label (p, instr);
		}

		/* XXX */
		if (instr->expr) {
			xlist_t *e = instr->expr;
			char c = *((char*)e->data);
			int i; /* = atoi ((char*)instr->expr->next->data); */
			Word data;
			if (!parse_string_data (p, instr->expr->next->data, &data, NULL)) {
				/* XXX */
				fatal (NULL, instr->fname, instr->line_num);
			}
			i = word_to_integer (&data);
			switch (c) {
			case '+':
				word_set (&instr->data, word_to_integer (&instr->data) + i);
				break;
			case '-':
				word_set (&instr->data, word_to_integer (&instr->data) - i);
				break;
			case '*':
				word_set (&instr->data, word_to_integer (&instr->data) * i);
				break;
			case '/':
				word_set (&instr->data, word_to_integer (&instr->data) / i);
				break;
			default:
				fatal (NULL, instr->fname, instr->line_num);
			}
		}

		if (instr->list) {
			xlist_t *tmp;
			for (tmp = instr->list; tmp != NULL; tmp = tmp->next) {
				Word *data = tmp->data;
				/* XXX: checkme - if DATA_TYPE_WORD, write LSB or MSB first? */
				dump_byte (p, cell, data->lsb);
				cell++;
				WORD_INC (&instr->addr);
				if (instr->dtype == DATA_TYPE_WORD) {
					/* col = dump_byte (p, instr->addr, data->msb, col, is_last); */
					dump_byte (p, cell, data->msb);
					cell++;
					WORD_INC (&instr->addr);
				}
			}
		}
		else if (!HAVE_OPCODE(instr)
		         || instr->mode == AM_IMMEDIATE
		         || instr->mode == AM_RELATIVE
		         || instr->mode == AM_INDIRECT_X
		         || instr->mode == AM_INDIRECT_Y
		         || instr->mode == AM_ZERO_PAGE
		         || instr->mode == AM_ZERO_PAGE_X
		         || instr->mode == AM_ZERO_PAGE_Y) {

			/* Byte */
			if (instr->immediate_msb) {
				dump_byte (p, cell, instr->data.msb);
				cell++;
			}
			else {
				dump_byte (p, cell, instr->data.lsb);
				cell++;
			}
		}
		else {

			/* Word, LSB first */
			dump_byte (p, cell, instr->data.lsb);
			cell++;
			dump_byte (p, cell, instr->data.msb);
		}

	}

	return col;
}


static void
write_instruction_asm (Parser *p, Instruction *instr)
{
	write_program_counter (p, &instr->addr);

	if (HAVE_OPCODE(instr)) {
		if (!show_counter) {
			write_char (p, '\t');
		}
		write_string (p, OP_GET_MNEMONIC(instr->opcode_id));
		if (instr->mode == AM_IMMEDIATE) {
			write_char (p, ' ');
			write_char (p, CHAR_IMMEDIATE);
		}
		else if (instr->mode == AM_ACCUMULATOR) {
			write_string (p, " A");
		}
		else if (instr->mode != AM_IMPLIED) {
			write_char (p, ' ');
		}
	}

	if (IS_BLOCK(instr)) {
		const char *s;
		s = (instr->dtype == DATA_TYPE_BYTE ? DIRECTIVE_BLKB : DIRECTIVE_BLKW);
		write_string (p, s);
		write_char (p, ' ');
		write_int (p, instr->block_size);
		write_string (p, ", ");
		if (instr->dtype == DATA_TYPE_BYTE) {
			write_byte (p, instr->data.lsb, TRUE);
		}
		else {
			write_word (p, &instr->data, TRUE);
		}
	}
	else if (!HAVE_OPCODE(instr)
	         || (instr->mode != AM_IMPLIED && instr->mode != AM_ACCUMULATOR)) {

		if (instr->label) {
			instruction_data_from_label (p, instr);
		}

		/* XXX */
		if (instr->expr) {
			xlist_t *e = instr->expr;
			char c = *((char*)e->data);
			int i; /* = atoi ((char*)instr->expr->next->data); */
			Word data;
			if (!parse_string_data (p, instr->expr->next->data, &data, NULL)) {
				/* XXX */
				fatal (NULL, instr->fname, instr->line_num);
			}
			i = word_to_integer (&data);
			switch (c) {
			case '+':
				word_set (&instr->data, word_to_integer (&instr->data) + i);
				break;
			case '-':
				word_set (&instr->data, word_to_integer (&instr->data) - i);
				break;
			case '*':
				word_set (&instr->data, word_to_integer (&instr->data) * i);
				break;
			case '/':
				word_set (&instr->data, word_to_integer (&instr->data) / i);
				break;
			default:
				fatal (NULL, instr->fname, instr->line_num);
			}
		}

		if (!HAVE_OPCODE(instr)) {
			/* This is a raw byte without a preceding opcode */
			write_string (p, DIRECTIVE_BYTE);
			write_char (p, ' ');
		}

		if (instr->list) {
			xlist_t *tmp;
			for (tmp = instr->list; tmp != NULL; tmp = tmp->next) {
				Word *data = tmp->data;
				/* XXX: checkme - if DATA_TYPE_WORD, write LSB or MSB first? */
				write_byte (p, data->lsb, TRUE);
				if (instr->dtype == DATA_TYPE_WORD) {
					write_string (p, ", ");
					write_byte (p, data->msb, TRUE);
				}
				if (tmp->next) {
					write_string (p, ", ");
				}
			}
		}
		else if (!HAVE_OPCODE(instr)
		         || instr->mode == AM_IMMEDIATE
		         || instr->mode == AM_RELATIVE
		         || instr->mode == AM_INDIRECT_X
		         || instr->mode == AM_INDIRECT_Y
		         || instr->mode == AM_ZERO_PAGE
		         || instr->mode == AM_ZERO_PAGE_X
		         || instr->mode == AM_ZERO_PAGE_Y) {

			/* Byte */

			if (instr->mode == AM_INDIRECT_X || instr->mode == AM_INDIRECT_Y) {
				write_char (p, '(');
			}

			if (instr->immediate_msb) {
				write_byte (p, instr->data.msb, TRUE);
			}
			else {
				write_byte (p, instr->data.lsb, TRUE);
			}

			if (instr->mode == AM_INDIRECT_X) {
				write_string (p, ",x)");
			}
			else if (instr->mode == AM_DIRECT_X
			         || instr->mode == AM_ZERO_PAGE_X) {
				write_string (p, ",x");
			}
			else if (instr->mode == AM_INDIRECT_Y) {
				write_string (p, "),y");
			}
			else if (instr->mode == AM_DIRECT_Y
			         || instr->mode == AM_ZERO_PAGE_Y) {
				write_string (p, ",y");
			}
		}
		else {

			/* Word */

			if (instr->mode == AM_INDIRECT) {
				write_char (p, '(');
			}

			write_word (p, &instr->data, TRUE);

			if (instr->mode == AM_DIRECT_X) {
				write_string (p, ",x");
			}
			else if (instr->mode == AM_DIRECT_Y) {
				write_string (p, ",y");
			}

			if (instr->mode == AM_INDIRECT) {
				write_char (p, ')');
			}

		}

	}

	write_char (p, '\n');
}


static int
write_instruction (Parser *p, Instruction *instr, int col, xbool_t is_last)
{
	switch (output_mode) {

		case OUTPUT_OBJ:
			write_instruction_obj (p, instr);
			break;

		case OUTPUT_HEX:
			col = write_instruction_hex (p, instr, col, is_last);
			break;

		case OUTPUT_ASM:
			write_instruction_asm (p, instr);
			break;
	}

	return col;
}


/* Check for unresolved symbols.  */

static void
check_labels (Parser *p)
{
	Label *label;
	xlist_t *this;
	char *msg;

	for (this = p->labels; this != NULL; this = this->next) {
		label = this->data;
#if 0
		if (strcasecmp (label->name, "A") == 0) {
			msg = xstr_dup_printf ("Assignment to reserved label 'A' (line %d, %s)",
			                       label->line_num, label->fname);
			xlist_append (&p->errors, msg);
		}
#endif
		if (!label->assigned) {
			msg = xstr_dup_printf ("Label '%s' is never assigned (Line %d, %s)",
			                       label->name, label->line_num, label->fname);
			xlist_append (&p->errors, msg);
		}
	}
}


/* Check for impossible addressing modes.  */

static void
check_instructions (Parser *p)
{
	Instruction *instr;
	xlist_t *this;
	char *msg;

	for (this = p->instructions; this != NULL; this = this->next) {
		instr = this->data;
		if (instr->opcode_id != -1) {
			/* There's an opcode, so check its addressing mode. */
			if (instr->mode == AM_NONE
				|| OP_GET_CODE(instr->opcode_id, instr->mode) == -1) {

				msg = xstr_dup_printf ("Unexpected addressing mode (%d) for '%s' (Line %d, %s)",
				                       instr->mode,
				                       opcode[instr->opcode_id].mnemonic,
				                       instr->line_num, instr->fname);

				xlist_append (&p->errors, msg);

			}
		}
	}
}


static void
compile (Parser *p)
{
	xlist_t *this;
	int col = 1;

	/* Check for unresolved symbols. */
	check_labels (p);

	/* Check for impossible addressing modes. */
	check_instructions (p);

	if (!xlist_is_empty (p->errors)) {
		exit (EXIT_FAILURE);
	}

	if (!quiet) {
		fprintf (stderr, "%s: %d labels, %d instructions\n", prog,
		         xlist_length (p->labels), xlist_length (p->instructions));
	}

	/* Write output file. */
	if (output_mode == OUTPUT_SYM) {
		Label *label;
		char *spacer;
		int spacer_len;
		for (this = p->labels; this != NULL; this = this->next) {
			label = this->data;
			spacer_len = 24 - strlen (label->name);
			if (spacer_len < 1) {
				spacer_len = 1;
			}
			spacer = xstr_nfill (spacer_len, ' ');
			fprintf (p->f_out, "%s%s= $%04X\n",
				label->name, spacer, word_to_integer (&label->data));
			xfree (spacer);
		}
		/* XXX: Should be able to write all output modes in a single run. */
		return;
	}

	for (this = p->instructions; this != NULL; this = this->next) {
		col = write_instruction (p, this->data, col, this->next == NULL);
	}

	if (output_mode == OUTPUT_HEX) {
		fprintf (p->f_out, "\n");
	}
}


/* Abort the program if too few tokens in this list. */

static void
check_required_length (Parser *p, xlist_t *token, int min_length)
{
	if (xlist_length (token) < min_length) {
		fatal (NULL, p->fname_in, p->line_num);
	}
}


/* Parse .org $XXXX */

static void
parse_org (Parser *p, xlist_t *this)
{
	Word data;

	check_required_length (p, this, 2);

	/* Skip to value. */
	this = this->next;

	if (!parse_string_data (p, TOK_TEXT(this), &data, NULL)) {
		string_data_error (p, TOK_TEXT(this));
	}

	word_copy (&p->pc, &data);

	/* Error if not end of tokens. */
	if (this->next) {
		fatal (NULL, p->fname_in, p->line_num);
	}
}


/* Parse FOOBAR .equ $XXXX
         FOOBAR = $XXXX  */

static void
parse_equ (Parser *p, xlist_t *this)
{
	Label *label;
	Word data;

	check_required_length (p, this, 3);

	if (!(label = get_label (p, TOK_TEXT(this), NULL))) {
		fatal (xstr_dup_printf ("Re-declaration of label %s", TOK_TEXT(this)),
		       p->fname_in, p->line_num);
	}

	/* Skip to value. */
	this = this->next->next;

	if (!parse_string_data (p, TOK_TEXT(this), &data, NULL)) {
		string_data_error (p, TOK_TEXT(this));
	}

	label_assign (p, label, &data);

	/* Error if not end of tokens. */
	if (this->next) {
		fatal (NULL, p->fname_in, p->line_num);
	}
}


/* Parse FOOBAR[:]  */

static xlist_t*
parse_label (Parser *p, xlist_t *this)
{
	Label *label;

	label = get_label (p, TOK_TEXT(this), &p->pc);
	assert (label != NULL);

	/* Skip to next token. */
	this = this->next;

	/* Skip the optional ":" if present. */
	if (this && strcmp (TOK_TEXT(this), LABEL_TERMINATOR) == 0) {
		this = this->next;
	}

	/* More tokens may follow. */

	return this;
}


/* Parse .byte $XX [,$XX [,...]]
         .word $XXXX [,$XXXX [,...]] */

static Instruction*
parse_db (Parser *p, xlist_t *this)
{
	Instruction *instr;
	Word *data;

	check_required_length (p, this, 2);
	instr = instruction_new (p, -1);
	instr->dtype = strcasecmp (TOK_TEXT(this), DIRECTIVE_BYTE) == 0 ?
	               DATA_TYPE_BYTE : DATA_TYPE_WORD;

	word_copy (&instr->addr, &p->pc);

	/* Skip to first value. */
	this = this->next;

	while (this) {

		if (*TOK_TEXT(this) == '"') {

			/* Expect a string - convert it to a list of bytes. */

			const char *q;

			if (instr->dtype != DATA_TYPE_BYTE) {
				fatal (xstr_dup_printf ("String data requires directive '%s'",
				       DIRECTIVE_BYTE), p->fname_in, p->line_num);
			}

			for (q = TOK_TEXT(this) + 1; *q != '\0'; q++) {
				data = xnew0 (Word, 1);
				word_set (data, *q);
				xlist_append (&instr->list, data);
				WORD_INC (&p->pc);
			}
		}
		else {

			/* Expect a numeric item. */

			data = xnew0 (Word, 1);

			if (!parse_string_data (p, TOK_TEXT(this), data, NULL)) {
				string_data_error (p, TOK_TEXT(this));
			}
			xlist_append (&instr->list, data);
			WORD_INC (&p->pc);
			if (instr->dtype == DATA_TYPE_WORD) {
				WORD_INC (&p->pc);
			}

		}

		this = this->next;

		if (this) {

			if (TOK_TYPE(this) == TOK_SYNTAX && *TOK_TEXT(this) == ',') {
				this = this->next;
			}

			if (!this) {
				fatal (xstr_dup ("Missing final element"),
				       p->fname_in, p->line_num);
			}
		}
	}

	return instr;
}


/* Parse .blkb SIZE [,FILL]
         .blkw SIZE [,FILL] */

static Instruction*
parse_blkb (Parser *p, xlist_t *this)
{
	Instruction *instr;
	Word data;
	int i;

	check_required_length (p, this, 2);
	instr = instruction_new (p, -1);
	instr->dtype = strcasecmp (TOK_TEXT(this), DIRECTIVE_BLKB) == 0 ?
	               DATA_TYPE_BYTE : DATA_TYPE_WORD;

	/* Skip to size. */
	this = this->next;

	if (!parse_string_data (p, TOK_TEXT(this), &data, NULL)) {
		string_data_error (p, TOK_TEXT(this));
	}

	instr->block_size = word_to_integer (&data);

	/* fprintf (stderr, "[%d]\n", instr->block_size); */

	word_copy (&instr->addr, &p->pc);

	for (i = 0; i < instr->block_size; i++) {
		WORD_INC (&p->pc);
		if (instr->dtype == DATA_TYPE_WORD) {
			WORD_INC (&p->pc);
		}
	}

	/* Skip to filler. */
	this = this->next;

	/* Fill-data given? */
	if (this && TOK_TYPE(this) == TOK_SYNTAX && *TOK_TEXT(this) == ',') {
		check_required_length (p, this, 2);
		this = this->next;
		if (!parse_string_data (p, TOK_TEXT(this), &instr->data, NULL)) {
			string_data_error (p, TOK_TEXT(this));
		}
	}
	else {
		/* Fill the block with zero by default. */
		word_set (&instr->data, 0);
	}

	/* Error if not end of tokens. */
	if (this && this->next) {
		fatal (NULL, p->fname_in, p->line_num);
	}

	return instr;
}


/* Parse .include "FILENAME.ASM"  */

static void
parse_incl (Parser *p, xlist_t *this)
{
	xlist_t *included;

	check_required_length (p, this, 2);

	/* Skip to filename. */
	this = this->next;

	if (!TOK_IS_STRING(this)) {
		fatal (xstr_dup_printf ("Unable to parse %s filename",
		       DIRECTIVE_INCLUDE), p->fname_in, p->line_num);
	}

	if ((included = load_source (p, TOK_TEXT(this) + 1)) != NULL) {
		xlist_join (&p->instructions, &included);
	}

	xlist_free (&included);
}


static Instruction*
parse_opcode (Parser *p, xlist_t* this)
{
	Instruction *instr;
	int id;

	id = opcode_id_from_mnemonic (TOK_TEXT(this));
	instr = instruction_new (p, id);
	instr->mode = (id == -1 ?
	               AM_NONE : opcode_default_address_mode (&opcode[id]));
	word_copy (&instr->addr, &p->pc);
	WORD_INC (&p->pc);

	return instr;
}


static void
token_list_free (xlist_t *list)
{
	xlist_t *this;

	for (this = list; this != NULL; this = this->next) {
		xfree (TOK_TEXT(this));
		xfree (this->data);
	}
	xlist_free (&list);
}


static Instruction*
parse_line (Parser *p, const char *text, xbool_t *stop)
{
	xlist_t     *list, *this;
	Token       *token;
	char        *s;
	int         start, end, length;
	TokenType   type;
	Label       *label;
	Instruction *instr;

#if 0
	immediate = FALSE;
#endif
	list = NULL;
	instr = NULL;
	start = 0;
	length = 0;

	while ((s = get_token (p, text, start, &end, &type, &length)) != NULL) {
		token = xnew (Token, 1);
		token->s = s;
		token->type = type;
		xlist_append (&list, token);
		start = end;
	}

	this = list;

	if (TOK_TYPE(this) == TOK_DIRECTIVE
		&& strcasecmp (TOK_TEXT(this), DIRECTIVE_END) == 0) {

		/* .end */

		token_list_free (list);
		*stop = TRUE;
		return NULL;

	}
	else if (TOK_TYPE(this) == TOK_DIRECTIVE
	    && strcasecmp (TOK_TEXT(this), DIRECTIVE_ORG) == 0) {

		/* .org $XXXX */

		parse_org (p, this);
		token_list_free (list);
		return NULL;

	}
	else if (TOK_TYPE(this) == TOK_DIRECTIVE
	         && strcasecmp (TOK_TEXT(this), DIRECTIVE_INCLUDE) == 0) {

		/* .include "FILENAME.asm" */

		parse_incl (p, this);
		token_list_free (list);
		return NULL;

	}
	else if (TOK_TYPE(this) == TOK_LABEL && this->next
	         && (strcasecmp (TOK_TEXT(this->next), DIRECTIVE_EQU) == 0
	             || strcmp (TOK_TEXT(this->next), "=") == 0)) {

		/* LABEL .equ $XXXX */
		/* LABEL = $XXXX */

		parse_equ (p, this);
		token_list_free (list);
		return NULL;

	}

	if (TOK_TYPE(this) == TOK_LABEL) {

		/* LABEL: */

		this = parse_label (p, this);
		if (this == NULL) {
			token_list_free (list);
			return NULL;
		}
	}

	if (TOK_TYPE(this) == TOK_DIRECTIVE
	    && (strcasecmp (TOK_TEXT(this), DIRECTIVE_BYTE) == 0
	        || strcasecmp (TOK_TEXT(this), DIRECTIVE_WORD) == 0)) {

		/* .byte $XX, $XX, $XX, ... */
		/* .word $XXXX, $XXXX, $XXXX, ... */

		instr = parse_db (p, this);
		token_list_free (list);

		return instr;

	}

	if (TOK_TYPE(this) == TOK_DIRECTIVE
	    && (strcasecmp (TOK_TEXT(this), DIRECTIVE_BLKB) == 0
	        || strcasecmp (TOK_TEXT(this), DIRECTIVE_BLKW) == 0)) {

		/* .blkb SIZE [,FILL] */
		/* .blkw SIZE [,FILL] */

		instr = parse_blkb (p, this);
		token_list_free (list);

		return instr;

	}

	if (this == NULL) {
		token_list_free (list);
		return NULL;
	}

	/* Expect an opcode... */

	if (TOK_TYPE(this) != TOK_OPCODE) {
		fatal (xstr_dup ("Illegal label or missing opcode"),
		       p->fname_in, p->line_num);
	}

	instr = parse_opcode (p, this);

	if (this->next == NULL) {
		char tmpstr[5];
		strncpy (tmpstr, TOK_TEXT(this), 4);
		token_list_free (list);
		if (instr->mode != AM_IMPLIED) {
			fatal (xstr_dup_printf ("Missing operand for '%s'", &tmpstr),
			       p->fname_in, p->line_num);
		}
		return instr;
	}

	this = this->next;

	/* Parse an operand... */

	instr->dtype = DATA_TYPE_WORD; /* Till we know otherwise. */

	if (TOK_TYPE(this) == TOK_SYNTAX
		&& *TOK_TEXT(this) == CHAR_LBRACKET) {

		/* ( DATA )
		   Indirect addressing. */

		instr->mode = AM_INDIRECT;
		this = this->next;

		if (this == NULL) {
			fatal (NULL, p->fname_in, p->line_num);
		}
	}

	if (TOK_TYPE(this) == TOK_SYNTAX && *TOK_TEXT(this) == CHAR_IMMEDIATE) {

		/* # Immediate addessing. */

		/* immediate = TRUE; */
		if (instr->mode == AM_INDIRECT) {
			fatal (xstr_dup ("Immediate mode addressing not allowed here"),
			       p->fname_in, p->line_num);
		}
		this = this->next;
		if (instr->mode != AM_RELATIVE) {
			instr->mode = AM_IMMEDIATE;
		}
		if (this == NULL) {
			fatal (NULL, p->fname_in, p->line_num);
		}
		instr->dtype = DATA_TYPE_BYTE;
		if (TOK_TYPE(this) == TOK_SYNTAX &&
			(*TOK_TEXT(this) == CHAR_MSB || *TOK_TEXT(this) == CHAR_LSB)) {
			if (*TOK_TEXT(this) == CHAR_MSB) {
				instr->immediate_msb = TRUE;
			}
			this = this->next;
			if (this == NULL) {
				fatal (NULL, p->fname_in, p->line_num);
			}
		}

		/* XXX: in-line arithmetic */

#if 1
		/* Expect end of line. */
		if (this->next != NULL) {
			fatal (NULL, p->fname_in, p->line_num);
		}
#endif

	}

	if (TOK_TYPE(this) == TOK_LABEL) {

		/* LABEL */

		/* "A" isn't really a label, but anyway... */
		if (strcasecmp (TOK_TEXT (this), STR_ACCUMULATOR) == 0
		    && op_is_shift (instr->opcode_id)) {
			instr->mode = AM_ACCUMULATOR;
		}
		else if (strcmp (TOK_TEXT (this), STR_PC) == 0) {
			word_copy (&instr->data, &instr->addr);
		}
		else if ((label = label_lookup_by_name (p, TOK_TEXT(this)))) {
			instr->label = label;
		}
		else {
			instr->label = get_label (p, TOK_TEXT(this), NULL);
		}

#if 0
		if (instr->mode != AM_IMMEDIATE
		    && instr->mode != AM_INDIRECT
		    && instr->mode != AM_RELATIVE
		    && word_to_integer (&instr->data) <= 0xFF) {

			instr->mode = AM_ZERO_PAGE;
		}
#endif

		/* XXX: implement in-line arithmetic properly */
		if (this->next && TOK_TYPE (this->next) == TOK_SYNTAX
		    && (*TOK_TEXT(this->next) == '+'
			    || *TOK_TEXT(this->next) == '-'
				|| *TOK_TEXT(this->next) == '*'
				|| *TOK_TEXT(this->next) == '/')) {

			this = this->next;
			xlist_append (&instr->expr, xstr_dup (TOK_TEXT(this)));

			this = this->next;
			xlist_append (&instr->expr, xstr_dup (TOK_TEXT(this)));

		}

#if 1
		/* Expect end of line, or ",", or ")". */
		if (this->next != NULL
		    && (strcmp (TOK_TEXT(this->next), ",") != 0
		        && strcmp (TOK_TEXT(this->next), STR_RBRACKET) != 0)) {
			fatal (NULL, p->fname_in, p->line_num);
		}
#endif

	}
	else if (TOK_TYPE(this) == TOK_DATA) {

		/* A raw byte or word. */

		xbool_t is_word;

		if (!parse_string_data (p, TOK_TEXT(this), &instr->data, &is_word)) {
			string_data_error (p, TOK_TEXT(this));
		}

		/* XXX: in-line arithmetic */

		if (instr->mode != AM_IMMEDIATE
			&& instr->mode != AM_INDIRECT
			&& instr->mode != AM_RELATIVE
			&& !is_word) {

			instr->mode = AM_ZERO_PAGE;
		}

#if 1
		/* Expect end of line, or ",", or ")". */
		if (this->next != NULL
		    && (strcmp (TOK_TEXT(this->next), ",") != 0
		        && strcmp (TOK_TEXT(this->next), STR_RBRACKET) != 0)) {
			fatal (NULL, p->fname_in, p->line_num);
		}
#endif

	}
	else {
		fatal (NULL, p->fname_in, p->line_num);
	}

	if (instr->mode != AM_ACCUMULATOR
		&& instr->mode != AM_IMMEDIATE
		&& instr->mode != AM_RELATIVE
		&& instr->mode != AM_INDIRECT
		&& instr->mode != AM_ZERO_PAGE) {

		WORD_INC (&p->pc);

	}
	else if (instr->mode == AM_INDIRECT
	         && opcode[instr->opcode_id].code[AM_INDIRECT] == OP_JMP_I) {

		/* JMP ($1234) */

		WORD_INC (&p->pc);
	}

	if (instr->mode != AM_ACCUMULATOR) {
		WORD_INC (&p->pc);
	}

	if (instr->mode == AM_INDIRECT) {

		this = this->next;
		if (this == NULL) {
			fatal (NULL, p->fname_in, p->line_num);
		}
		if (*TOK_TEXT(this) == CHAR_RBRACKET) {
			if (this->next == NULL) {

				/* JMP ($1234) */

				token_list_free (list);
				return instr;
			}

			/* Indirect Y : ($34),Y
			   This is handled below, when we get to parsing ,Y */

		}

	}

	if (instr->mode == AM_INDIRECT
	    && TOK_TYPE(this) == TOK_SYNTAX && strcmp (TOK_TEXT(this), ",") == 0) {

		/* (NN,X)
		   Indexed indirect addressing. */

		check_required_length (p, this, 3);
		instr->mode = AM_INDIRECT_X;
		this = this->next;

		if (strcasecmp (TOK_TEXT(this), "Y") == 0) {
			fatal (xstr_dup ("Can't do pre-indexed indirect addressing with Y"),
			       p->fname_in, p->line_num);
		}
		else if (strcasecmp (TOK_TEXT(this), "X") != 0) {
			fatal (xstr_dup ("Expected index X"), p->fname_in, p->line_num);
		}
		this = this->next;
		if (*TOK_TEXT(this) != CHAR_RBRACKET) {
			fatal (NULL, p->fname_in, p->line_num);
		}
		this = this->next;
		token_list_free (list);
		if (this != NULL) {
			fatal (NULL, p->fname_in, p->line_num);
		}

		return instr;
	}

	if (!(this = this->next)) {
		token_list_free (list);
		return instr;
	}

	if (TOK_TYPE(this) == TOK_SYNTAX && strcmp (TOK_TEXT(this), ",") == 0) {

		/* ,X  ,Y
		   Indexed direct addressing, or indexed indirect. */

		check_required_length (p, this, 2);
		this = this->next;
		if (strcasecmp (TOK_TEXT(this), "X") == 0) {
			if (instr->mode == AM_INDIRECT) {
				fatal (xstr_dup ("Can't do post-indexed indirect addressing with X"),
				       p->fname_in, p->line_num);
			}
			else if (instr->mode == AM_ZERO_PAGE) {
				instr->mode = AM_ZERO_PAGE_X;
			}
			else {
				instr->mode = AM_DIRECT_X;
			}
		}
		else if (strcasecmp (TOK_TEXT(this), "Y") == 0) {
			if (instr->mode == AM_INDIRECT) {
				instr->mode = AM_INDIRECT_Y;
			}
			else if (instr->mode == AM_ZERO_PAGE) {
				instr->mode = AM_ZERO_PAGE_Y;
			}
			else {
				instr->mode = AM_DIRECT_Y;
			}
		}
		else {
			fatal (xstr_dup ("Expected index X or Y"),
			       p->fname_in, p->line_num);
		}
	}

	/* If any tokens remain, something's wrong... */
	if (this->next) {
		fatal (xstr_dup ("Unexpected token"), p->fname_in, p->line_num);
	}

	token_list_free (list);

	return instr;
}


/* Set the current input file name.  */

static void
set_fname_in (Parser *p, const char *fname)
{
	xfree (p->fname_in);
	p->fname_in = xstr_dup (fname);
}


/* Return TRUE if the named file is already open. */

static xbool_t
already_open (Parser *p, const char *fname)
{
	Include *incl;
	xlist_t *this;

	for (this = p->includes; this != NULL; this = this->next) {
		incl = this->data;
		if (strcmp (incl->fname, fname) == 0) {
			return TRUE;
		}
	}

	return FALSE;
}


/* Register and open an included source file.  */

static Include*
new_source_file (Parser *p, const char *fname)
{
	FILE *f_in;
	Include *incl;

	if (already_open (p, fname)) {
		fatal (xstr_dup_printf ("Recursive inclusion of file '%s'", fname),
		       NULL, -1);
	}

	if ((f_in = fopen (fname, "r")) == NULL) {
		fprintf (stderr, "%s: %s: %s\n", prog, fname, strerror (errno));
		exit (EXIT_FAILURE);
	}

	incl = xnew (Include, 1);
	incl->fname = xstr_dup (fname);
	incl->f_in = f_in;
	incl->parent_line_num = p->line_num;

	/* Add it to the list of open files and update the current fname_in. */
	xlist_push (&p->includes, incl);
	set_fname_in (p, incl->fname);

	return incl;
}


/* Close the current source file and return to processing its parent.  */

static void
close_source (Parser *p, Include *incl)
{
	int line_num;

	line_num = incl->parent_line_num;

	if (incl->f_in != NULL) {
		fclose (incl->f_in);
	}
	xfree (incl->fname);
	xfree (incl);

	/* Update current fname_in to head of list. */
	if (!xlist_is_empty (p->includes)) {
		set_fname_in (p, ((Include*)p->includes->data)->fname);
		p->line_num = line_num;
	}
}


/* Parse a source file.  */

static xlist_t*
load_source (Parser *p, const char *fname)
{
	size_t      bufsize;
	ssize_t     result;
	char        *s, *buf = NULL;
	xlist_t     *list = NULL;
	Instruction *instr;
	Include     *incl;
	xbool_t     stop = FALSE;

	if (!quiet) {
		fprintf (stderr, "%s: %s '%s'\n", prog, DIRECTIVE_INCLUDE, fname);
	}

	incl = new_source_file (p, fname);
	p->line_num = 1;

	while ((result = xfile_getline (&buf, &bufsize, incl->f_in)) > 0) {

		s = xstr_strip (xstr_dup (buf));

		if (*s != '\0' && *s != CHAR_COMMENT) {

			if ((instr = parse_line (p, s, &stop)) != NULL) {
				xlist_append (&list, instr);
			}
			if (stop) {
				if (!quiet) {
					fprintf (stderr, "%s: %s '%s'\n",
					         prog, DIRECTIVE_END, p->fname_in);
				}
				break;
			}
		}

		xfree (s);

		p->line_num++;
	}

	xfree(buf);

	close_source (p, xlist_pop (&p->includes));

	return list;
}


static void
cleanup (void)
{
	xlist_t *tmp;

	if (parser->errors) {

		/* Complain. */
		for (tmp = parser->errors; tmp != NULL; tmp = tmp->next) {
			char *msg = tmp->data;
			fprintf (stderr, "%s: ERROR: %s\n", prog, msg);
			xfree (msg);
		}
		xlist_free (&parser->errors);
	}


	/* Close output file, unless it's stdout. */
	if (parser->f_out && parser->f_out != stdout) {
		fclose (parser->f_out);
	}

	/* Close any currently open source files. */
	while (!xlist_is_empty (parser->includes)) {
		close_source (parser, xlist_pop (&parser->includes));
	}
	xlist_free (&parser->includes);


	/* Free labels. */
	for (tmp = parser->labels; tmp != NULL; tmp = tmp->next) {
		label_free (tmp->data);
	}
	xlist_free (&parser->labels);


	/* Free instructions. */
	for (tmp = parser->instructions; tmp != NULL; tmp = tmp->next) {
		instruction_free (tmp->data);
	}
	xlist_free (&parser->instructions);


	/* Parser */
	xfree (parser->fname_in);
	xfree (parser->fname_out);
	xfree (parser);

	xfree (sep_addr);
	xfree (sep_col);

	xlist_pool_free ();

	xfree (prog);
}


static void
show_usage (FILE *f)
{
	fprintf (f, "\nUsage: %s [OPTIONS] FILE.asm\n\n", prog);

	fprintf (f, "   A simple featureless 6502 machine code assembler.\n\n");

	fprintf (f, "OPTIONS:\n\n");

	fprintf (f, "  -o,  --output FILE  Write to 'FILE' (or '-' for stdout)\n");
	fprintf (f, "  -or, --origin ADDR  Use ADDR as the base address (default: 0x0000)\n");
	fprintf (f, "  -c,  --counter      Prefix each line with its memory address\n");
	fprintf (f, "  -sa, --saddress XX  Follow the address counter with XX ('' to disable)\n");
	fprintf (f, "  -l,  --lower        Write opcodes and hexadecimals in lower-case\n");
	fprintf (f, "  -u,  --upper        Write opcodes and hexadecimals in upper-case\n");
	fprintf (f, "  -q,  --quiet        Suppress progress messages\n");
	fprintf (f, "  -v,  --version      Show program version information\n");
	fprintf (f, "  -h,  --help         Show this usage message\n");

	fprintf (f, "\nOutput modes:\n\n");

	fprintf (f, "  -obj, --obj         Output raw bytes\n");
	fprintf (f, "  -hex, --hex         Output hexadecimal\n");
	fprintf (f, "  -asm, --asm         Output assembly language\n");
	fprintf (f, "  -sym, --symtab      Output a symbol table\n");

	fprintf (f, "\nIn hexadecimal output mode:\n\n");

	fprintf (f, "  -sc,  --scolumn XX  Write XX between columns ('' to disable)\n");
	fprintf (f, "  -w,   --width N     Wrap output at column N (default: %d)\n",
	         DEFAULT_DUMP_WIDTH);

	fprintf (f, "\nSeparators (XX) can be up to two characters long.\n");
	fprintf (f, "The sequence \\t (backslash t) represents one TAB character.\n");

	fprintf (f, "\n");
}


static void
show_version (FILE *f)
{
	fprintf (f, "%s version %s", prog, VERSION);
	fprintf (f, ", bug reports to %s <%s>\n", AUTHOR_NAME, AUTHOR_EMAIL);
}


int
main (int argc, char *argv[])
{
	int opt, idx, o;
	const char *fname_in;

	atexit (cleanup);

	prog = xfile_basename (argv[0]);

	xlist_pool_init (0);

	parser = xnew0 (Parser, 1);
	sep_col = xstr_dup (SEPARATOR_COL);
	sep_addr = xstr_dup (SEPARATOR_ADDR);

	opt = 1;
	idx = 1;

	while (idx < argc && opt > XGETOPT_NONE) {

		opt = xgetopt (argc, argv, &idx, option, N_OPTS);

		switch (opt) {

		case OPT_HELP:
			show_usage (stdout);
			exit (EXIT_SUCCESS);
		case OPT_OBJ:
			output_mode = OUTPUT_OBJ;
			break;
		case OPT_HEX:
			output_mode = OUTPUT_HEX;
			break;
		case OPT_ASM:
			output_mode = OUTPUT_ASM;
			break;
		case OPT_SYM:
			output_mode = OUTPUT_SYM;
			break;
		case OPT_COUNTER:
			show_counter = TRUE;
			break;
		case OPT_ORIGIN:
			o = parse_ascii_integer (argv[idx++]);
			word_set (&parser->pc, o);
			break;
		case OPT_SEPARATOR_ADDR:
			xfree (sep_addr);
			sep_addr = unescape_tabs (xstr_dup (argv[idx]));
			if (strlen (sep_addr) > 2) {
				fatal (xstr_dup ("Separators are limited to two characters in length"),
				       NULL, -1);
			}
			idx++;
			break;
		case OPT_SEPARATOR_COL:
			xfree (sep_col);
			sep_col = unescape_tabs (xstr_dup (argv[idx]));
			if (strlen (sep_col) > 2) {
				fatal (xstr_dup ("Separators are limited to two characters in length"),
				       NULL, -1);
			}
			idx++;
			break;
		case OPT_OUTPUT:
			xfree (parser->fname_out);
			parser->fname_out = xstr_dup (argv[idx++]);
			break;
		case OPT_WIDTH:
			dump_width = atoi (argv[idx++]);
			break;
		case OPT_LOWER:
			uppercase_output = FALSE;
			break;
		case OPT_UPPER:
			uppercase_output = TRUE;
			break;
		case OPT_QUIET:
			quiet = TRUE;
			break;
		case OPT_VERSION:
			show_version (stdout);
			exit (EXIT_SUCCESS);
		case XGETOPT_MISSING_ARG:
			fprintf (stderr, "%s: Option '%s' requires an argument\n",
				prog, argv[idx - 1]);
			exit (EXIT_FAILURE);
		case XGETOPT_UNKNOWN:
			fprintf (stderr, "%s: Unknown option '%s'\n", prog, argv[idx - 1]);
			exit (EXIT_FAILURE);
		default:
			break;

		}

	}

	if (argc - idx != 1) {
		fprintf (stderr, "%s: Try '--help' for usage information\n", prog);
		exit (EXIT_FAILURE);
	}

	fname_in = argv[idx];

	if (parser->fname_out == NULL) {
		const char *ext;
		switch (output_mode) {
		case OUTPUT_HEX:
			ext = FNAME_EXT_HEX;
			break;
		case OUTPUT_OBJ:
			ext = FNAME_EXT_OBJ;
			break;
		case OUTPUT_ASM:
			ext = FNAME_EXT_ASM;
			break;
		case OUTPUT_SYM:
			ext = FNAME_EXT_SYM;
			break;
		}
		parser->fname_out = xstr_dup_printf ("%s.%s", FNAME_OUT, ext);
	}

	/* Define the special "*" label. */
	parser->label_pc = get_label (parser, STR_PC, &parser->pc);

	parser->instructions = load_source (parser, fname_in);

	if (strcmp (parser->fname_out, "-") == 0) {
		parser->f_out = stdout;
	}
	else if ((parser->f_out = fopen (parser->fname_out, "w")) == NULL) {
		fprintf (stderr, "%s: %s: %s\n",
		         prog, parser->fname_out, strerror (errno));
		exit (EXIT_FAILURE);
	}

	compile (parser);

	return EXIT_SUCCESS;
}

