/* emu.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <strings.h>
#include <time.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xlist.h"
#include "xlib/xfile.h"
#include "xlib/xgetopt.h"

#include "config.h"
#include "memory.h"
#include "machine.h"
#include "support.h"

#define DEFAULT_MACHINE   "machines" DIR_SEPARATOR_S "default"
#define DEFAULT_RAM_SIZE  0xFFFF

#define AUTHOR            "Tim Musson"
#define AUTHOR_EMAIL      "trmusson@gmail.com"
#ifndef VERSION
#define VERSION           "sweet_amnesia"
#endif


static char *progname = NULL;

/* List of files to load into simulated RAM. */
static xlist_t *load = NULL;

Machine *machine = NULL;


static XGetOption option[] = {
	{1, FALSE, "fullscreen"},
	{1, TRUE,  "load"},
	{1, TRUE,  "machine"},
	{1, TRUE,  "ram"},
	{1, TRUE,  "startup"},
	{1, TRUE,  "tape"},
	{1, FALSE, "quiet"},
	{1, FALSE, "version"},
	{1, FALSE, "help"}
};


enum {
	OPT_FULLSCREEN,
	OPT_LOAD,
	OPT_MACHINE,
	OPT_RAM,
	OPT_STARTUP,
	OPT_TAPE,
	OPT_QUIET,
	OPT_VERSION,
	OPT_HELP,
	N_OPTS
};


static void
show_version (void)
{
	printf ("%s %s\n", progname, VERSION);
	exit (EXIT_SUCCESS);
}


static void
show_help (void)
{
	printf ("\nUsage: %s [OPTIONS]\n\n", progname);
	printf ("  Simulation of 6502-based microcomputers, Ohio Scientific and related\n");
	printf ("  machines in particular.\n\n");
	printf ("Options:\n\n");
	printf ("  -f,  --fullscreen           Start in fullscreen mode (F11 to toggle).\n");
	printf ("  -l,  --load FILE:ADDR       Load a file of hex values into RAM at ADDR.\n");
	printf ("                              The --load option can appear multiple times.\n");
	printf ("  -m,  --machine DIRECTORY    Simulate the computer described by the profile\n");
	printf ("                              stored in this directory.\n");
	printf ("  -q,  --quiet                Only write to stderr if problems encountered.\n");
	printf ("  -r,  --ram SIZE             Memory size, in bytes.\n");
	printf ("  -s,  --startup ADDR         Begin execution at ADDR on startup.\n");
	printf ("  -t,  --tape FILE            Use FILE as a simulated audio cassette.\n");
	printf ("  -v,  --version              Show program version.\n");
	printf ("  -h,  --help                 Show this usage message.\n");
	printf ("\nNumeric values can be given in decimal or hexidecimal, with hexidecimal\n");
	printf ("values prefixed with '0x' (example: 0xFF00).\n");
	printf ("\nBug reports to %s <%s>\n\n", AUTHOR, AUTHOR_EMAIL);

	exit (EXIT_SUCCESS);
}


static void
cleanup (void)
{
	machine_free (machine);

	xlist_foreach (load, (XList_f)xfree, NULL);
	xlist_free (&load);

	xlist_pool_free ();
	xfree (progname);
}


int
main (int argc, char *argv[])
{
	const char *profile;
	const char *fname_tape = NULL;
	int ram_size = DEFAULT_RAM_SIZE;
	int startup_addr = -1;
	xbool_t quiet = FALSE;
	xbool_t fullscreen = FALSE;
	int opt, idx;


	progname = xfile_basename (argv[0]);
	profile = DEFAULT_MACHINE;

	atexit (cleanup);

	xlist_pool_init (0);
	srand (time (NULL));

	opt = 1;
	idx = 1;

	while (idx < argc && opt > XGETOPT_NONE) {

		opt = xgetopt (argc, argv, &idx, option, N_OPTS);

		switch (opt) {

		case OPT_FULLSCREEN:
			fullscreen = TRUE;
			break;
		case OPT_LOAD:
			xlist_append (&load, xstr_dup (argv[idx++]));
			break;
		case OPT_MACHINE:
			profile = argv[idx++];
			break;
		case OPT_RAM:
			ram_size = parse_ascii_integer (argv[idx++]);
			break;
		case OPT_STARTUP:
			startup_addr = parse_ascii_integer (argv[idx++]);
			break;
		case OPT_TAPE:
			fname_tape = argv[idx++];
			break;
		case OPT_QUIET:
			quiet = TRUE;
			break;
		case OPT_HELP:
			show_help ();
			break;
		case OPT_VERSION:
			show_version ();
			break;

		}

	}

	if ((machine = machine_new (progname, profile, ram_size,
		load, fname_tape, quiet, fullscreen)) != NULL) {

		if (machine_start (machine, startup_addr)) {
			exit (EXIT_SUCCESS);
		}

	}

	return (EXIT_FAILURE);
}

